#include "wrap/thread.h"
#include "wrap/synch.h"
#include <stdio.h>
#include <stdlib.h>

#define PETERSONS 0
#define BUSYFLAG  1
#define NOOP      2
#define MUTEX     3
#define INTR      4
#define XCHNG     5

#ifndef V
// Default choice of algorithm if none explicitly specified.
#define V PETERSONS
#endif

#ifndef V
#error "Define V as PETERSONS, BUSYFLAG, NOOP, MUTEX, INTR, or XCHNG"
#endif

void plock(int);
void punlock(int);

#if V == PETERSONS /* * * * * * * * * * * PET's * * * * * * * * * * */

#define ME(id)    (id)
#define OTHER(id) (!id)

bool i_want_in[2] = { false, false };
bool whose_turn; /* set in plock, always before use */

/* Peterson's solution.
 *
 * Presumption: Only two threads.
 *
 * Switch at note 1 or 2:
 *  First to write whose_turn will win the lock
 *  Last to write whose_turn will wait
 *
 * Switch at note 3:
 *  If turn becomes mine other will wait until I do not want in anymore
 *  If other want in he'll set turn to mine and have to wait
 *
 * If I'm the only one that want in I'll get in
 * If I wait I'll continue when other get out
 */
void plock(int tid) {
	i_want_in[ME(tid)] = true;
	/* note 1 */
	whose_turn = OTHER(tid);
	/* note 2 */
	while ( i_want_in[OTHER(tid)] && /* note 3 */
		whose_turn == OTHER(tid) )
		; /* busy wait */
}

void punlock(int tid) {
	i_want_in[ME(tid)] = false;
}

#elif V == BUSYFLAG /* * * * * * * * * * * Busy-flag * * * * * * * * * * */

bool busy = false;
int owner = -1;

void plock(int tid) {
	while ( busy )
		; /* busy wait */
	busy = true;
	owner = tid;
}

void punlock(int tid) {
	owner = -1;
	busy = false;
}

#elif V == NOOP /* * * * * * * * * * * No op * * * * * * * * * * */

void plock(int tid) {}
void punlock(int tid) {}

#elif V == MUTEX /* * * * * * * * * Pthread mutex * * * * * * * * */

extern struct lock seqlock;
void plock(int tid) { lock_acquire(&seqlock); }
void punlock(int tid) { lock_release(&seqlock); }

#elif V == INTR /* * * * * * * * * Interrupt disable * * * * * * * * */

void plock(int tid) { asm volatile ("cli"); }
void punlock(int tid) { asm volatile("sti"); }

#elif V == XCHNG /* * * * * * * * * Atomic swap * * * * * * * * */

volatile int busy = false;
int owner = -1;

void plock(int tid) {
	while (atomic_compare_and_swap(&busy, 0, 1) == 1)
		;

	/* asm volatile ("loop: " */
    /*             "movl %2, %%eax;" */
    /*             "xchgl %%eax, %0;" */
    /*             "test %%eax, %%eax;" */
    /*             "jnz loop;" */
    /*             : "=m" (busy) */
    /*             : "m" (busy), "r" (true) */
    /*             : "%eax" */
	/* 	); */

	owner = tid;
}

void punlock(int tid) {
	owner = -1;
	busy = 0;
}

#endif

struct lock seqlock;
int seq = 0;

static void seq_0(void *input) {
	int tid = 0;

	for ( ; ; ) {
		plock(tid);
		fprintf(stdout, "%d Peterson was a clever guy.\n", ++seq);
		punlock(tid);
	}

	sema_up(input);
}

static void seq_1(void *input) {
	int tid = 1;

	for ( ; ; )
	{
		plock(tid);
		fprintf(stdout, "%d His solution may actually work!\n", ++seq);
		punlock(tid);
	}

	sema_up(input);
}

int main(void) {
	struct semaphore zero, one;

	lock_init(&seqlock);
	sema_init(&zero, 0);
	sema_init(&one, 1);

	thread_create("", 0, &seq_0, &zero);
	thread_create("", 0, &seq_1, &one);

	sema_down(&zero);
	sema_down(&one);

	return 0;
}
